      program ringomi

c     July 2000  Kelly Chance
c     Modified from previous Ring programs for OMI use.

c     This program calculates the Ring effect for pure Raman scattering of
c     the Fraunhofer spectrum plus the contribution from interference by
c     terrestrial absorption (from a "trace gas"). Ideally, terrestrial
c     absorption would be matched to the particular atmospheric path. To allow
c     for more generality, this program calculates the optically thin term as
c     well as the next term in the expansion for the Beer-Lambert law (i.e.,
c     correction for saturation in the terrestrial absorption:
c     exp (-tau) = 1-tau+tau**2/2 + ....).
c     Ringspec, Ring2, and Ring3 are the Fraunhofer only, 1st terrestrial
c     correction, and 2nd terrestrial correction. Provision is made for
c     trimming the ends of the spectrum to account for the convolution with
c     the slit function. For GOME calculations, for example, an extra 5 nm on
c     either side of the desired Ring spectrum should be trimmed off. DOAS-
c     type Ring spectra, which are the real Ring spectra divided by the
c     Fraunhofer spectrum (and sometimes logarithmized, but be careful of
c     negative values in the terrestrial correction terms; the correct
c     expansion of the logarithm of the ratio of backscatter plus a linear
c     Ring correction to the solar irradiance, used for DOAS fitting, gives
c     Ring / irradiance as the first term, with higher terms normally
c     ignored) and/or high pass filtered (usually), can be generated by
c     selecting the appropriate options described below. Orthogonal
c     eigenfunctions can be generated and the whole lot can be normalized to
c     RMS values of 1. Renormalizing is both to avoid numerical problems and to
c     output spectra with convenient scales for spectral fitting. If absolute
c     Ring cross sections are desired, the user should set if_renorm = .false.
c     The program includes convolution with a Gaussian slit function specified
c     by its half-width at 1/e intensity (hw1e). If the full width at
c     half-maximum (fwhm) is preferred, note that
c     fwhm = 2 * sqrt (ln 2) * hw1e = 1.665109 * hw1e.
c     For other choices of slit width, a high-resolution spectrum may be
c     convolved with the appropriate slit function afterward. Current
c     dimensioning is for 40000 grid points, sufficient for the OMI range with
c     a Fraunhofer reference spectrum at 0.01 nm resolution. This may need to
c     be increased for larger spectral ranges.

c     f77 -o ringomi.x ringomi.f
c     ringomi.x << EOF
c     ring.inp
c     ring.out
c     EOF

c     Files:

c     input 1 - Two rows of input parameters, followed by the Fraunhofer
c     spectrum, in 2-column ascii. The Ring spectrum will be output on this
c     grid, to within the limits imposed by trimming. The first row contains
c     the temperature for calculation of the  Raman spectra, the number of
c     grid points for extending the Raman spectrum convolution (+- slit
c     +- 200 cm-1 Ring contributions), the hw1e width for the slit function
c     convolution, and the lower and upper limits for trimming of the output.
c     The second row contains the logical controls.
c     Logicals:
c     if_trace - whether to include a trace gas spectrum to correct for
c               interference to Raman scatteric by atmospheric absorption
c               (e.g., by O3 or NO2).
c     if_trim - whether to trim to the limits given by trimlo and trimhi.
c     if_renorm - whether to renormalize (I suggest yes for most fitting
c                 applications).
c     if_div -  whether to divide by the solar spectrum ("DOAS").
c     if_hpf -  whether to high pass filter, by subtracting a cubic
c               polynomial ("DOAS").
c     if_log -  whether to take the logarithm of the divided spectrum ("DOAS").
c     if_eigen - whether to output orthogonal eigenfunctions instead of raw
c               Ring spectra (either for normal fitting or DOAS).
c               Orthogonalization is performed after trimming. Note that if the
c               region of the spectrum used for fitting changes, the
c               eigenvectors will no longer by orthogonal. However,
c               re-orthogonalization is easy; in any case, orthogonalization
c               or re-orthogonalization are not strictly necessary, since they
c               consist of taking appropriate linear combinations of the Ring
c               spectra. Fitting with a set of functions or with linear
c               combinations of the functions should give exactly the same
c               result.

c     input 2 - The optional terrestrial interfering spectrum (usually NO2 or
c               O3), also in 2-column ascii, read in if (if_trace). This will
c               be splined to the grid of the Fraunhofer spectrum, so make sure
c               it overlaps (extrapolation is usually a bad idea).

c     output -  The position grid, followed by either the Ring spectra or the
c               eigenvectors for the Ring spectra, if that option is selected.
c               If (if_div), the one (or three) columns of true Ring spectra
c               will be followed by the respective columns of "differential"
c               Ring spectra. For solar input in the usual units of W m-2 nm-1,
c               the output will be the Ring source per scatterer in W nm-1; the
c               Ring divided by solar will be in m**2.

C     Please communicate any corrections, additions, or improvements to me:

c     Kelly Chance
c     Harvard-Smithsonian Center for Astrophysics
c     60 Garden Street, Cambridge, MA 02138 USA
c     phone/fax +1-617-495-7389
c     kchance@cfa.harvard.edu
c     http://cfa-www.harvard.edu/~kchance

      implicit real*8 (a - h, o - z)
      character*40 input1, input2, output
      dimension pos (40000), solar (40000), ptrace (40000),
     * trace (40000), d2trace (40000), ringspec (40000), ring2 (40000),
     * ring3 (40000), ringspec_hpf (40000), ring2_hpf (40000),
     * ring3_hpf (40000), solarmod (40000), tracemod (40000),
     * vec1 (40000), vec2 (40000), vec3 (40000), vec4 (40000),
     * vec5 (40000), vec6 (40000)
      real*8 n2pos (48), n2xsec (48), n2xref (48)
      real*8 o2pos (185), o2xsec (185), o2xref (185)
      dimension params (4), lista (4), covar (4, 4)
      logical if_trace, if_trim, if_renorm, if_div, if_hpf, if_log,
     * if_eigen

      data n2pos /
     * -194.3015, -186.4226, -178.5374, -170.6459, -162.7484,
     * -154.8453, -146.9368, -139.0233, -131.1049, -123.1819,
     * -115.2547, -107.3235, -99.3886, -91.4502, -83.5086,
     * -75.5642, -67.6171, -59.6676, -51.7162, -43.7629,
     * -35.8080, -27.8519, -19.8950, -11.9373, 11.9373,
     * 19.8950, 27.8519, 35.8080, 43.7629, 51.7162,
     * 59.6676, 67.6171, 75.5642, 83.5086, 91.4502,
     * 99.3886, 107.3235, 115.2547, 123.1819, 131.1049,
     * 139.0233, 146.9368, 154.8453, 162.7484, 170.6459,
     * 178.5374, 186.4226, 194.3015 /

      data o2pos /
     * -185.5861, -185.5690, -185.5512, -174.3154, -174.2980,
     * -174.2802, -163.0162, -162.9980, -162.9809, -151.6906,
     * -151.6730, -151.6551, -140.3405, -140.3230, -140.3046,
     * -128.9677, -128.9490, -128.9314, -117.5740, -117.5560,
     * -117.5372, -108.2632, -106.1613, -106.1430, -106.1240,
     * -104.3008, -96.8136, -94.7318, -94.7120, -94.6934,
     * -92.8515, -85.3489, -83.2871, -83.2671, -83.2473,
     * -81.3868, -73.8694, -71.8294, -71.8080, -71.7876,
     * -69.9077, -62.3771, -60.3611, -60.3373, -60.3157,
     * -58.4156, -50.8730, -48.8851, -48.8571, -48.8332,
     * -46.9116, -40.1158, -39.3565, -37.4069, -37.3688,
     * -37.3406, -35.3953, -27.8241, -25.9472, -25.8745,
     * -25.8364, -25.8125, -23.8629, -16.2529, -16.2529,
     * -14.3761, -14.3033, -14.1686, -12.2918, -2.1392,
     * -2.1202, -2.1016, -2.0843, -2.0843, -2.0818,
     * -2.0614, -2.0398, -2.0159, -2.0116, -1.9877,
     * -1.9735, -1.9496, -1.9455, -1.9217, -1.9003,
     * -1.8803, -1.8768, -1.8605, -1.8422, -0.1347,
     * 0.1347, 1.8422, 1.8605, 1.8768, 1.8803,
     * 1.9003, 1.9217, 1.9455, 1.9496, 1.9735,
     * 1.9877, 2.0116, 2.0159, 2.0398, 2.0614,
     * 2.0818, 2.0843, 2.0843, 2.1016, 2.1202,
     * 2.1392, 12.2918, 14.1686, 14.3033, 14.3761,
     * 16.2529, 16.2529, 23.8629, 25.8125, 25.8364,
     * 25.8745, 25.9472, 27.8241, 35.3953, 37.3406,
     * 37.3688, 37.4069, 39.3565, 40.1158, 46.9116,
     * 48.8332, 48.8571, 48.8851, 50.8730, 58.4156,
     * 60.3157, 60.3373, 60.3611, 62.3771, 69.9077,
     * 71.7876, 71.8080, 71.8294, 73.8694, 81.3868,
     * 83.2473, 83.2671, 83.2871, 85.3489, 92.8515,
     * 94.6934, 94.7120, 94.7318, 96.8136, 104.3008,
     * 106.1240, 106.1430, 106.1613, 108.2632, 115.7318,
     * 117.5372, 117.5560, 117.5740, 119.6952, 128.9314,
     * 128.9490, 128.9677, 140.3046, 140.3230, 140.3405,
     * 151.6551, 151.6730, 151.6906, 162.9809, 162.9980,
     * 163.0162, 174.2802, 174.2980, 174.3154, 185.5512,
     * 185.5690, 185.5861, 196.7919, 196.8100, 196.8269 /

      write (*, '(5x, a)') 'Enter Fraunhofer input file.'
      read (*, 10) input1
10    format (a)
      open (unit = 1, file = input1, status = 'old')

      do i = 1, 40000
        ringspec (i) = 0.d0
        ring2 (i) = 0.d0
        ring3 (i) = 0.d0
      end do

c     Read the temperature (temp = 250 K is the approximate average
c     temperature for scattering); npix for endpoint calculations
c     (+- slit +- 200 cm-1 Ring contributions) - this is the number of
c     points at our Fraunhofer resolution of usually 0.01 nm needed to
c     calculate the slit out to a low value (I like 0.001) plus the
c     equivalent of 200 cm-1 at the long-wavelength end of the Ring
c     spectrum being calculated; the hw1e width for smearing by the
c     slit, dldomi, and the wavelengths for final trimming, trimlo
c     and trimhi.
      read (1, *) temp, npix, dldomi, trimlo, trimhi

c     Read the logical controls:
      read (1, *) if_trace, if_trim, if_renorm, if_div, if_hpf, if_log,
     * if_eigen
      write (*, *) ' if_trace = ', if_trace
      write (*, *) ' if_trim = ', if_trim
      write (*, *) ' if_renorm = ', if_renorm
      write (*, *) ' if_div = ', if_div
      write (*, *) ' if_hpf = ', if_hpf
      write (*, *) ' if_log = ', if_log
      write (*, *) ' if_eigen = ', if_eigen

      if (if_trace) then
        write (*, '(5x, a)') 'Enter trace gas input file.'
        read (*, 10) input2
        open (unit = 2, file = input2, status = 'old')
      end if
      write (*, '(5x, a)') 'Enter output file.'
      read (*, 10) output
      open (unit = 3, file = output, status = 'unknown')

c     Set up the rotational Raman spectra
c     These later need to be multiplied by gamma**2/lambda**4
      call raman_n2 (temp, n2xref)
      call raman_o2 (temp, o2xref)

c     Read the solar spectrum data, usually in W m-2 nm-1.
      i = 1
20    read (1, *, end = 30) pos (i), solar (i)
        i = i + 1
        go to 20
30    npoints = i - 1
      write (*, *) ' npoints = ', npoints
      dldpx = (pos (npoints) - pos (1)) / (npoints - 1.)
      write (*, *) ' dldpx = ', dldpx

c     Read the trace gas cross sections. For consistency, they should be in
c     m2, but it is not vitally important for fitting purposes.
      if (if_trace) then
        tracesum = 0.d0
        i = 1
40      read (2, *, end = 50) ptrace (i), trace (i)
          i = i + 1
          go to 40
50      ntrace = i - 1
        write (*, *) ' ntrace = ', ntrace

c       Spline the trace gas cross sections to the Fraunhofer grid.
        call spline (ptrace, trace, ntrace, d2trace)
        call splint (ptrace, trace, d2trace, ntrace, npoints, pos,
     *  tracemod)
      end if

c     Add up Ring contributions over wavelengths and lines; remember that
c     the change in photon energy is opposite that of the molecular energy.
      do i = 1, npoints
c       1.d30 * gamma (m-3).
        sigsq = 1.d6 / (pos (i)**2)
        gamman2 = -0.601466 + 238.557 / (186.099 - sigsq)
        gammao2 = 0.07149 + 45.9364 / (48.2716 - sigsq)
        do j = 1, 48
          sigprime = (1.d7 / pos (i)) - n2pos (j)
c         n2xsec (j) is now the total cross section for this state
c         for scattering of input radiation having wavelength pos (i):
          n2xsec (j) = n2xref (j) * (sigprime**4) * (gamman2**2)
c         Contributions from +- npix pixels away
          do k = max (1, i - npix), min (npoints, i + npix)
            contrib =  n2xsec (j) *
     *      slit (dldomi, pos (i), - n2pos (j), pos (k), dldpx)
            ringspec (k) = ringspec (k) + solar (i) * contrib
            if (if_trace) then
              ring2 (k) = ring2 (k) + solar (i) * tracemod (i) * contrib
              ring3 (k) = ring3 (k) + solar (i) * tracemod (i)**2 *
     *        contrib
            end if
          end do
        end do
        do j = 1, 185
          sigprime = (1.d7 / pos (i)) - o2pos (j)
c         o2xsec (j) is now the total cross section for this state
c         for scattering of input radiation having wavelength pos (i):
          o2xsec (j) = o2xref (j) * (sigprime**4) * (gammao2**2)
c         Contributions from +- npix pixels away
          do k = max (1, i - npix), min (npoints, i + npix)
            contrib =  o2xsec (j) *
     *      slit (dldomi, pos (i), - o2pos (j), pos (k), dldpx)
            ringspec (k) = ringspec (k) + solar (i) * contrib
            if (if_trace) then
              ring2 (k) = ring2 (k) + tracemod (i) * contrib
              ring3 (k) = ring3 (k) + tracemod (i)**2 * contrib
            end if
          end do
        end do
      end do

c     The following is old code, left in to remind us that if we were not
c     renormalizing anywhere, we need to multiply by d-52 to get the correct
c     units for the absolute cross sections in m-2: We calculate gamman2
c     and gammao2 in d-30 m3 and then calculate n2xsec and o2xsec values
c     as (gamma**2) * (sigprime**4), where sigprime is in cm-1 instead of
c     m-1, giving an extra factor of d8.
c     do i = 1, npoints
c       ringspec (i) = ringspec (i) * 1.d-30 * 1.d-30 * 1.d8
c       ring2 (i) = ring2 (i) * 1.d-30 * 1.d-30 * 1.d8
c       ring3 (i) = ring3 (i) * 1.d-30 * 1.d-30 * 1.d8
c     end do

c     If "differential" versions of the cross sections are calculated, we
c     convolve the solar spectrum down before trimming, to avoid introducing
c     endpoint problems.
      if (if_div) call gauss (pos, solar, solarmod, npoints, dldomi)

c     Apply the trimming, first locating the points to keep.
      if (if_trim) then
        do i = 1, npoints
          if (pos (i) .ge. trimlo) then
            ntrimlo = i
            go to 60
          end if
        end do
60      continue
        do i = npoints, 1, -1
          if (pos (i) .lt. trimhi) then
            ntrimhi = i + 1
            go to 70
          end if
        end do
70      ntrim = ntrimhi - ntrimlo + 1
        write (*, *) ntrimlo, ntrimhi, ntrim
        do i = 1, ntrim
          j = i + ntrimlo - 1
          ringspec (i) = ringspec (j)
          if (if_trace) then
            ring2 (i) = ring2 (j)
            ring3 (i) = ring3 (j)
          end if
          solarmod (i) = solarmod (j)
          pos (i) = pos (j)
        end do
        npoints = ntrim
      end if

c     After trimming: renormalize the Ring spectra, if (if_renorm)
      if (if_renorm) then
        call normalize (ringspec, ringspec, npoints)
        if (if_trace) then
          call normalize (ring2, ring2, npoints)
          call normalize (ring3, ring3, npoints)
        end if
        if (if_div) call normalize (solarmod, solarmod, npoints)
      end if

c     Divide by solar spectrum as part of "differential" process.
      if (if_div) then
        do i = 1, npoints
          ringspec_hpf (i) = ringspec (i) / solarmod (i)
          if (if_trace) then
            ring2_hpf (i) = ring2 (i) / solarmod (i)
            ring3_hpf (i) = ring3 (i) / solarmod (i)
          end if
        end do
      end if

c     Take logarithms as part of "differential" process, against my advice.
      if (if_log) then
        do i = 1, npoints
          ringspec_hpf (i) = dlog10 (ringspec_hpf (i))
          if (if_trace) then
            ring2_hpf (i) = dlog10 (ring2_hpf (i))
            ring3_hpf (i) = dlog10 (ring3_hpf (i))
          end if
        end do
      end if

c     Subtract cubic polynomial as part of "differential" process.
      if (if_hpf) then
        do i = 1, 4
          lista (i) = i
        end do
        call lfit (pos, ringspec_hpf, npoints, params, lista, covar,
     *  chisq)
        do i = 1, npoints
          ringspec_hpf (i) = ringspec_hpf (i) - (params (1) +
     *    params (2) * pos (i) + params (3) * pos (i) * pos (i) +
     *    params (4) * ((pos (i))**3))
        end do
        if (if_trace) then
          call lfit (pos, ring2_hpf, npoints, params, lista, covar,
     *    chisq)
          do i = 1, npoints
            ring2_hpf (i) = ring2_hpf (i) - (params (1) +
     *      params (2) * pos (i) + params (3) * pos (i) * pos (i) +
     *      params (4) * ((pos (i))**3))
          end do
          call lfit (pos, ring3_hpf, npoints, params, lista, covar,
     *    chisq)
          do i = 1, npoints
            ring3_hpf (i) = ring3_hpf (i) - (params (1) +
     *      params (2) * pos (i) + params (3) * pos (i) * pos (i) +
     *      params (4) * ((pos (i))**3))
          end do
        end if
      end if

c     Eigenvector analysis, by Gram-Schmidt orthogonalization, with optional
c     normalization; this is only applicable when if_trace = .true.
      if (if_eigen) then
        if (.not. if_trace) then
	  write (*, *) ' Error: eigenvector analysis without trace gas.'
          go to 80
        end if
c       Eigenvectors for normal Ring spectra
        if (if_renorm) call normalize (ringspec, vec1, npoints)
        call overlap (vec1, ring2, npoints, del12)
        do i = 1, npoints
          vec2 (i) = - del12 * vec1 (i) + ring2 (i)
        end do
        if (if_renorm) call normalize (vec2, vec2, npoints)
        call overlap (vec1, ring3, npoints, del13)
        call overlap (vec2, ring3, npoints, del23)
        do i = 1, npoints
          vec3 (i) = - del13 * vec1 (i) - del23 * vec2 (i) + ring3 (i)
        end do
        if (if_renorm) call normalize (vec3, vec3, npoints)
c       Eigenvectors for "differential" Ring spectra
        if (if_div) then
          if (if_renorm) call normalize (ringspec_hpf, vec4, npoints)
          call overlap (vec4, ring2_hpf, npoints, del45)
          do i = 1, npoints
            vec5 (i) = - del45 * vec4 (i) + ring2_hpf (i)
          end do
          if (if_renorm) call normalize (vec5, vec5, npoints)
          call overlap (vec4, ring3_hpf, npoints, del46)
          call overlap (vec5, ring3_hpf, npoints, del56)
          do i = 1, npoints
            vec6 (i) = - del46 * vec4 (i) - del56 * vec5 (i) +
     *      ring3_hpf (i)
          end do
          if (if_renorm) call normalize (vec6, vec6, npoints)
        end if  ! on if_div
      else if (if_renorm .and. if_div) then
c       Normalize "differential" spectra again (to rms = 1), for
c       non-eigenvector case (real Ring spectra have not changed).
        call normalize (ringspec_hpf, ringspec_hpf, npoints)
        call normalize (ring2_hpf, ring2_hpf, npoints)
        call normalize (ring3_hpf, ring3_hpf, npoints)
      end if  ! on if_eigen

c     Output results
      renorm = 1.d-52
      if (if_renorm) renorm = 1.d0
      if (if_trace) then
c       Write either 3 or 6 columns (depending on if_div) of either Ring
c       spectra or their orthogonal eigenvectors.
        if (if_eigen) then
          if (if_div) then
            do i = 1, npoints
              write (3, '(f10.5, 1p6e13.5)') pos (i), renorm * vec1 (i),
     *        renorm * vec2 (i), renorm * vec3 (i), renorm * vec4 (i),
     *        renorm * vec5 (i), renorm * vec6 (i)
            end do
          else
            do i = 1, npoints
              write (3, '(f10.5, 1p3e13.5)') pos (i), renorm * vec1 (i),
     *        renorm * vec2 (i), renorm * vec3 (i)
            end do
          end if
        else
c       Write either 3 or 6 columns (depending on if_div) of either Ring
c       spectra or their orthogonal eigenvectors.
          if (if_div) then
            do i = 1, npoints
              write (3, '(f10.5, 1p6e13.5)') pos (i), renorm *
     *        ringspec (i), renorm * ring2 (i), renorm * ring3 (i),
     *        renorm * ringspec_hpf (i), renorm * ring2_hpf (i),
     *        renorm * ring3_hpf (i)
            end do
          else
            do i = 1, npoints
              write (3, '(f10.5, 1p3e13.5)') pos (i), renorm *
     *        ringspec (i), renorm * ring2 (i), renorm * ring3 (i)
            end do
          end if
        end if
      else
c       Write either 2 or 3 columns (depending on if_div) of Ring spectra.
        if (if_div) then
          do i = 1, npoints
            write (3, '(f10.5, 1p2e13.5)') pos (i), renorm *
     *      ringspec (i), renorm * ringspec_hpf (i)
          end do
        else
          do i = 1, npoints
            write (3, '(f10.5, 1pe13.5)') pos (i), renorm * ringspec (i)
          end do
        end if
      end if

80    close (unit = 1)
      if (if_trace) close (unit = 2)
      close (unit = 3)
      stop
      end

      function slit (dldomi, pnm1, pcm, pnm2, dldpx)

c     Use a Gaussian slit with 1/e halfwidth of dldomi, which would normally
c     be the omi resolution, in nm.

c     pnm1 is the position of the Fraunhofer source, in nm.
c     pcm is the (-) relative position of the Raman line, so that pprime,
c     defined below, is the location of the Raman-shifted line contributing
c     to the Ring spectrum.
c     pnm2 is the position of the contribution we are calculating with
c     this call.

c     It is important to remember here that we have calculated cross sections
c     Q (m-2) and what we really need are cross sections per nanometer, i.e.,
c     convolved with a slit function normalized to 1 nm-1. This is accomplished
c     here by deriving slitsum as the sum over samples (without including the
c     increment in wavelength) and sampling over the slit function in the calls
c     to this function.

c     There is a sampling problem which is not (yet) addressed here: The
c     Raman lines are quite narrow (0.0006-0.0009 nm halfwidths) in the
c     OMI range. If the spectrum is not broadened sufficiently, it is possible
c     to miss lines entirely. If it is insufficiently broadened, then they will
c     be poorly-sampled and the Ring spectrum will be quantitatively incorrect.
c     A binning option needs to be added to accommodate this. We normally use
c     a Fraunhofer spectrum at 0.01 nm resolution (DF). Errors of the scale of
c     DF/slitwidth can be expected in individual samples.

      implicit real*8 (a - h, o - z)
      logical first_call /.true./
      save first_call

c     calculate the slit normalization out to 0.001 times the center value.
      if (first_call) then
        write (*, *) ' slit normalization'
        first_call = .false.
        slitsum = 1.d0
        do i = 1, 1000
          slitadd = exp (- (dldpx * i)**2)
          slitsum = slitsum + 2.d0 * slitadd
          if (slitadd .le. 0.001) then
c           write (*, *) ' nslit = ', i
            go to 10
          end if
          if (i .eq. 1000) write (*, *) ' slitsum range failed.'
        end do
10      continue
      end if

c     lambda'
      sigprime = pcm + 1.d7 / pnm1
      pprime = 1.d7 / sigprime

c     Distance in pixels
      distance = (pprime - pnm2) / dldomi
      del = distance**2

      slit = dexp (- del) / slitsum

      return
      end

      subroutine gauss (pos, spec, specmod, npoints, hw1e)

c     Convolves input spectrum with Gaussian slit function of specified
c     HW1E (half-width at 1/e intensity).

      implicit real*8 (a - h, o - z)
      parameter (maxpts = 40000)
      dimension pos (maxpts), spec (maxpts)
      dimension specmod (maxpts), slit (maxpts)

      if (hw1e .eq. 0) then
        write (*, *) ' No Gaussian convolution applied.'
        do i = 1, npoints
          specmod (i) = spec (i)
        end do
        return
      end if

      emult = - 1.d0 / (hw1e**2)
      delpos = pos (2) - pos (1)

c     Apply slit function convolution: Calculate slit function values out
c     to 0.001 times x0 value, normalize so that sum = 1.
      slitsum = 1.d0
      slit0 = 1.d0
      do i = 1, 1000
        slit (i) = exp (emult * (delpos * i)**2)
        slitsum = slitsum + 2.d0 * slit (i)
        if (slit (i) / slit0 .le. 0.001) then
          nslit = i
          go to 40
        end if
      end do
40    continue
      slit0 = slit0 / slitsum
      do i = 1, nslit
        slit (i) = slit (i) / slitsum
      end do

c     Convolve spectrum.  reflect at endpoints.
      do i = 1, npoints
        specmod (i) = slit0 * spec (i)
        do j = 1, nslit
          nlo = i - j
          nhi = i + j
          if (nlo .lt. 0) nlo = - nlo
          if (nlo .eq. 0) nlo = 1
          if (nhi .gt. npoints) nhi = 2 * npoints - nhi
          specmod (i) = specmod (i) +
     *    slit (j) * (spec (nlo) + spec (nhi))
        end do
      end do

      return
      end

      subroutine normalize (vin, vout, npoints)

c     Normalize to rms = 1

      implicit real*8 (a - h, o - z)
      dimension vin (40000), vout (40000)

      const = 1.d0  ! overflow protection
      if (dabs (vin (1)) .ge. 1.d17) const = 1.d17
      sum = 0.d0
      do i = 1, npoints
        sum = sum + (vin (i) / const)**2
      end do
      rms = const * dsqrt (sum / npoints)
      write (*, *) ' normalize: sum, npoints, rms = ',
     * sum, npoints, rms
      do i = 1, npoints
        vout (i) = vin (i) / rms
      end do

      return
      end

      subroutine overlap (v1, v2, npoints, del)
      implicit real*8 (a - h, o - z)
      dimension v1 (40000), v2 (40000)

      del = 0.d0
      do i = 1, npoints
        del = del + v1 (i) * v2 (i)
      end do
      del = del / npoints

      return
      end

      subroutine raman_n2 (temp, n2xsec)

      implicit real*8 (a - h, o - z)
c     Statistical parameters, for partition function calculation
      real*8 n2stat_1 (31), n2stat_2 (31), n2stat_3 (31)
c     Rotational Raman line-specific parameters
      real*8 n2term (48), n2plactel (48), n2deg (48), n2nuc (48)
      real*8 n2frac (48), n2xsec (48)

      data n2stat_1 / 0., 1., 2., 3., 4., 5., 6., 7., 8.,
     * 9., 10., 11., 12., 13., 14., 15., 16., 17., 18., 19.,
     * 20., 21., 22., 23., 24., 25., 26., 27., 28., 29., 30. /

      data n2stat_2 / 6., 3., 6., 3., 6., 3., 6., 3., 6.,
     * 3., 6., 3., 6., 3., 6., 3., 6., 3., 6., 3.,
     * 6., 3., 6., 3., 6., 3., 6., 3., 6., 3., 6. /

      data n2stat_3 / 0.0000, 3.9791, 11.9373, 23.8741, 39.7892,
     * 59.6821, 83.5521, 111.3983, 143.2197, 179.0154,
     * 218.7839, 262.5240, 310.2341, 361.9126, 417.5576,
     * 477.1673, 540.7395, 608.2722, 679.7628, 755.2090,
     * 834.6081, 917.9574, 1005.2540, 1096.4948, 1191.6766,
     * 1290.7963, 1393.8503, 1500.8350, 1611.7467, 1726.5816,
     * 1845.3358 /

      data n2term /
     * 1290.7963, 1191.6766, 1096.4948, 1005.2540, 917.9574,
     * 834.6081, 755.2090, 679.7628, 608.2722, 540.7395,
     * 477.1673, 417.5576, 361.9126, 310.2341, 262.5240,
     * 218.7839, 179.0154, 143.2197, 111.3983, 83.5521,
     * 59.6821, 39.7892, 23.8741, 11.9373, 0.0000,
     * 3.9791, 11.9373, 23.8741, 39.7892, 59.6821,
     * 83.5521, 111.3983, 143.2197, 179.0154, 218.7839,
     * 262.5240, 310.2341, 361.9126, 417.5576, 477.1673,
     * 540.7395, 608.2722, 679.7628, 755.2090, 834.6081,
     * 917.9574, 1005.2540, 1096.4948 /

      data n2plactel /
     * 3.601d-1, 3.595d-1, 3.589d-1, 3.581d-1, 3.573d-1,
     * 3.565d-1, 3.555d-1, 3.544d-1, 3.532d-1, 3.519d-1,
     * 3.504d-1, 3.487d-1, 3.467d-1, 3.443d-1, 3.416d-1,
     * 3.383d-1, 3.344d-1, 3.294d-1, 3.231d-1, 3.147d-1,
     * 3.030d-1, 2.857d-1, 2.571d-1, 2.000d-1, 1.000d+0,
     * 6.000d-1, 5.143d-1, 4.762d-1, 4.545d-1, 4.406d-1,
     * 4.308d-1, 4.235d-1, 4.180d-1, 4.135d-1, 4.099d-1,
     * 4.070d-1, 4.044d-1, 4.023d-1, 4.004d-1, 3.988d-1,
     * 3.974d-1, 3.961d-1, 3.950d-1, 3.940d-1, 3.931d-1,
     * 3.922d-1, 3.915d-1, 3.908d-1 /

      data n2deg / 25., 24., 23., 22., 21., 20., 19., 18., 17., 16.,
     * 15., 14., 13., 12., 11., 10., 9., 8., 7., 6.,
     * 5., 4., 3., 2., 0., 1., 2., 3., 4., 5.,
     * 6., 7., 8., 9., 10., 11., 12., 13., 14., 15.,
     * 16., 17., 18., 19., 20., 21., 22., 23. /

      data n2nuc /
     * 3., 6., 3., 6., 3., 6., 3., 6., 3., 6., 3., 6., 3., 6.,
     * 3., 6., 3., 6., 3., 6., 3., 6., 3., 6., 6., 3., 6., 3., 6.,
     * 3., 6., 3., 6., 3., 6., 3., 6., 3., 6., 3., 6., 3., 6., 3.,
     * 6., 3., 6., 3. /

      pi = 3.14159265358979
      c2 = 1.438769
      emult = - c2 / temp

c     Calculate partition function
      qn2 = 0.
      do i = 1, 31
        qn2 = qn2 + (2. * n2stat_1 (i) + 1.) * n2stat_2 (i) *
     *  exp (emult * n2stat_3 (i))
      end do

      prefix = 256. * (pi**5) * 0.79 / 27.

c     Calculate population fractions for rotational Raman lines and
c     the cross sections, except for gamma**2/lambda**4
      do i = 1, 48
        n2frac (i) = (2. * n2deg (i) + 1.) * n2nuc (i) *
     *  exp (emult * n2term (i)) / qn2
        n2xsec (i) = prefix * n2frac (i) * n2plactel (i)
      end do

      return
      end

      subroutine raman_o2 (temp, o2xsec)

      implicit real*8 (a - h, o - z)
c     Statistical parameters, for partition function calculation
      real*8 o2stat_1 (54), o2stat_2 (54)
c     Rotational Raman line-specific parameters
      real*8 o2term (185), o2plactel (185), o2deg (185)
      real*8 o2frac (185), o2xsec (185)

      data o2stat_1 / 0., 2., 1., 2.,
     * 4., 3., 4., 6., 5., 8., 6., 7., 10., 8.,
     * 9., 12., 10., 11., 14., 12., 13., 16., 14., 15.,
     * 18., 16., 17., 20., 18., 19., 22., 20., 21., 24.,
     * 22., 23., 26., 24., 25., 28., 26., 27., 30., 28.,
     * 29., 32., 30., 31., 34., 32., 33., 36., 34., 35. /

      data o2stat_2 /
     * 0.0000, 2.0843, 3.9611, 16.2529, 16.3876, 18.3372,
     * 42.2001, 42.2240, 44.2117, 79.5646, 79.6070, 81.5805,
     * 128.3978, 128.4921, 130.4376, 188.7135, 188.8532, 190.7749,
     * 260.5011, 260.6826, 262.5829, 343.7484, 343.9697, 345.8500,
     * 438.4418, 438.7015, 440.5620, 544.5658, 544.8628, 546.7050,
     * 662.1030, 662.4368, 664.2610, 791.0344, 791.4045, 793.2100,
     * 931.3390, 931.7450, 933.5330, 1082.9941, 1083.4356,
     * 1085.2060, 1245.9750, 1246.4518, 1248.2040, 1420.2552,
     * 1420.7672, 1422.5020, 1605.8064, 1606.3533, 1608.0710,
     * 1802.5983, 1803.1802, 1804.8810 /

      data o2term /
     * 1606.3533, 1608.0710, 1605.8064, 1420.7672, 1422.5020,
     * 1420.2552, 1246.4518, 1248.2040, 1245.9750, 1083.4356,
     * 1085.2060, 1082.9941, 931.7450, 933.5330, 931.3390,
     * 791.4045, 793.2100, 791.0344, 662.4368, 664.2610,
     * 662.1030, 546.7050, 544.8628, 546.7050, 544.5658,
     * 544.8628, 440.5620, 438.7015, 440.5620, 438.4418,
     * 438.7015, 345.8500, 343.9697, 345.8500, 343.7484,
     * 343.9697, 262.5829, 260.6826, 262.5829, 260.5011,
     * 260.6826, 190.7749, 188.8532, 190.7749, 188.7135,
     * 188.8532, 130.4376, 128.4921, 130.4376, 128.3978,
     * 128.4921, 42.2001, 81.5805, 79.6070, 81.5805,
     * 79.5646, 79.6070, 44.2117, 42.2001, 44.2117,
     * 42.2240, 42.2001, 42.2001, 16.2529, 18.3372,
     * 18.3372, 16.3876, 16.2529, 16.2529, 546.7050,
     * 440.5620, 345.8500, 2.0843, 18.3372, 262.5829,
     * 190.7749, 130.4376, 81.5805, 44.2117, 44.2117,
     * 81.5805, 18.3372, 130.4376, 190.7749, 262.5829,
     * 345.8500, 3.9611, 440.5620, 546.7050, 16.3876,
     * 16.2529, 544.8628, 438.7015, 2.0843, 343.9697,
     * 260.6826, 188.8532, 128.4921, 16.3876, 79.6070,
     * 42.2240, 42.2001, 79.5646, 128.3978, 188.7135,
     * 260.5011, 0.0000, 16.2529, 343.7484, 438.4418,
     * 544.5658, 3.9611, 2.0843, 2.0843, 3.9611,
     * 0.0000, 2.0843, 18.3372, 16.3876, 16.3876,
     * 18.3372, 16.2529, 16.3876, 44.2117, 42.2240,
     * 44.2117, 42.2001, 42.2240, 2.0843, 81.5805,
     * 79.5646, 81.5805, 79.6070, 79.5646, 130.4376,
     * 128.3978, 130.4376, 128.4921, 128.3978, 190.7749,
     * 188.7135, 190.7749, 188.8532, 188.7135, 262.5829,
     * 260.5011, 262.5829, 260.6826, 260.5011, 345.8500,
     * 343.7484, 345.8500, 343.9697, 343.7484, 440.5620,
     * 438.4418, 440.5620, 438.7015, 438.4418, 546.7050,
     * 544.5658, 546.7050, 544.8628, 544.5658, 662.1030,
     * 664.2610, 662.4368, 791.0344, 793.2100, 791.4045,
     * 931.3390, 933.5330, 931.7450, 1082.9941, 1085.2060,
     * 1083.4356, 1245.9750, 1248.2040, 1246.4518, 1420.2552,
     * 1422.5020, 1420.7672, 1605.8064, 1608.0710, 1606.3533 /

      data o2plactel /
     * 3.630d-1, 3.630d-1, 3.637d-1, 3.622d-1, 3.622d-1,
     * 3.630d-1, 3.613d-1, 3.613d-1, 3.622d-1, 3.602d-1,
     * 3.602d-1, 3.612d-1, 3.589d-1, 3.589d-1, 3.601d-1,
     * 3.574d-1, 3.574d-1, 3.589d-1, 3.556d-1, 3.556d-1,
     * 3.573d-1, 2.079d-3, 3.533d-1, 3.534d-1, 3.555d-1,
     * 2.191d-3, 2.597d-3, 3.505d-1, 3.506d-1, 3.532d-1,
     * 2.755d-3, 3.337d-3, 3.468d-1, 3.471d-1, 3.504d-1,
     * 3.567d-3, 4.444d-3, 3.418d-1, 3.422d-1, 3.467d-1,
     * 4.800d-3, 6.211d-3, 3.348d-1, 3.354d-1, 3.416d-1,
     * 6.803d-3, 9.288d-3, 3.220d-1, 3.251d-1, 3.344d-1,
     * 1.127d-2, 1.062d-3, 1.387d-2, 3.013d-1, 3.077d-1,
     * 3.223d-1, 1.979d-2, 2.613d-2, 2.544d-1, 2.727d-1,
     * 3.020d-1, 1.476d-3, 4.342d-2, 9.234d-2, 6.596d-2,
     * 1.714d-1, 2.571d-1, 1.843d-2, 1.615d-1, 1.970d-3,
     * 2.445d-3, 3.116d-3, 1.077d-1, 7.690d-2, 4.105d-3,
     * 5.652d-3, 8.271d-3, 1.222d-2, 2.842d-2, 2.207d-2,
     * 1.470d-2, 5.132d-2, 8.256d-3, 5.647d-3, 4.103d-3,
     * 3.115d-3, 2.308d-1, 2.445d-3, 1.970d-3, 2.116d-3,
     * 3.810d-3, 2.076d-3, 2.593d-3, 1.385d-1, 3.329d-3,
     * 4.431d-3, 6.184d-3, 9.227d-3, 3.991d-2, 1.696d-2,
     * 1.867d-2, 3.474d-2, 1.078d-2, 7.483d-3, 5.200d-3,
     * 3.822d-3, 5.383d-1, 1.077d-1, 2.927d-3, 2.313d-3,
     * 1.874d-3, 2.692d-1, 1.843d-2, 4.628d-1, 4.000d-1,
     * 4.617d-1, 9.234d-2, 5.583d-2, 1.476d-3, 4.362d-1,
     * 4.286d-1, 4.579d-1, 3.193d-2, 2.339d-2, 4.214d-1,
     * 4.196d-1, 4.352d-1, 1.600d-2, 1.911d-3, 1.278d-2,
     * 4.130d-1, 4.118d-1, 4.210d-1, 1.038d-2, 7.519d-3,
     * 4.067d-1, 4.060d-1, 4.135d-1, 6.803d-3, 5.217d-3,
     * 4.021d-1, 4.017d-1, 4.070d-1, 4.800d-3, 3.831d-3,
     * 3.987d-1, 3.985d-1, 4.023d-1, 3.567d-3, 2.933d-3,
     * 3.961d-1, 3.959d-1, 3.988d-1, 2.755d-3, 2.317d-3,
     * 3.939d-1, 3.938d-1, 3.961d-1, 2.191d-3, 1.876d-3,
     * 3.922d-1, 3.921d-1, 3.940d-1, 1.785d-3, 3.908d-1,
     * 3.907d-1, 3.922d-1, 3.895d-1, 3.895d-1, 3.908d-1,
     * 3.885d-1, 3.885d-1, 3.896d-1, 3.876d-1, 3.876d-1,
     * 3.885d-1, 3.868d-1, 3.868d-1, 3.876d-1, 3.861d-1,
     * 3.861d-1, 3.868d-1, 3.855d-1, 3.855d-1, 3.861d-1 /

      data o2deg /
     * 32., 33., 34., 30., 31., 32., 28., 29., 30., 26., 27.,
     * 28., 24., 25., 26., 22., 23., 24., 20., 21., 22., 19., 18.,
     * 19., 20., 18., 17., 16., 17., 18., 16., 15., 14., 15., 16.,
     * 14., 13., 12., 13., 14., 12., 11., 10., 11., 12., 10., 9., 8.,
     * 9., 10., 8., 4., 7., 6., 7., 8., 6., 5., 4., 5., 6., 4., 4.,
     * 2., 3., 3., 4., 2., 2., 19., 17., 15., 2., 3., 13., 11., 9.,
     * 7., 5., 5., 7., 3., 9., 11., 13., 15., 1., 17., 19., 4., 2.,
     * 18., 16., 2., 14., 12., 10., 8., 4., 6., 6., 4., 8., 10., 12.,
     * 14., 0., 2., 16., 18., 20., 1., 2., 2., 1., 0., 2., 3., 4., 4.,
     * 3., 2., 4., 5., 6., 5., 4., 6., 2., 7., 8., 7., 6., 8., 9.,
     * 10., 9., 8., 10., 11., 12., 11., 10., 12., 13., 14., 13., 12.,
     * 14., 15., 16., 15., 14., 16., 17., 18., 17., 16., 18., 19.,
     * 20., 19., 18., 20., 22., 21., 20., 24., 23., 22., 26., 25.,
     * 24., 28., 27., 26., 30., 29., 28., 32., 31., 30., 34., 33.,
     * 32. /

      pi = 3.14159265358979
      c2 = 1.438769
      emult = - c2 / temp

c     Calculate partition function
      qo2 = 0.
      do i = 1, 54
        qo2 = qo2 + (2. * o2stat_1 (i) + 1.) *
     * exp (emult * o2stat_2 (i))
      end do

      prefix = 256. * (pi**5) * 0.21 / 27.

c     Calculate population fractions for rotational Raman lines and
c     the cross sections, except for gamma**2/lambda**4
      do i = 1, 185
        o2frac (i) = (2. * o2deg (i) + 1.) *
     *  exp (emult * o2term (i)) / qo2
        o2xsec (i) = prefix * o2frac (i) * o2plactel (i)
      end do

      return
      end

      subroutine lfit (x, y, ndata, a, lista, covar, chisq)

      implicit real*8 (a - h, o - z)
      dimension x (40000), y (40000), a (4),
     * lista (4), covar (4, 4), beta (4), afunc (4)

c     Dimension for the exact number of parameters:
      ncvm = 4
      mfit = 4
      ma = 4

      kk=mfit+1
      do j=1,ma
        ihit=0
        do k=1,mfit
          if (lista(k).eq.j) ihit=ihit+1
        enddo
        if (ihit.eq.0) then
          lista(kk)=j
          kk=kk+1
        else if (ihit.gt.1) then
          pause 'improper set in lista'
        endif
      enddo
      if (kk.ne.(ma+1)) pause 'improper set in lista'
      do j=1,mfit
        do k=1,mfit
          covar(j,k)=0.
        enddo
        beta(j)=0.
      enddo
      do i=1,ndata
        afunc (1) = 1.
        afunc (2) = x (i)
        afunc (3) = x (i) * x (i)
        afunc (4) = (x (i))**3
        ym=y(i)
        if(mfit.lt.ma) then
          do j=mfit+1,ma
            ym=ym-a(lista(j))*afunc(lista(j))
          enddo
        endif
        do j=1,mfit
          wt=afunc(lista(j))
          do k=1,j
            covar(j,k)=covar(j,k)+wt*afunc(lista(k))
          enddo
          beta(j)=beta(j)+ym*wt
        enddo
      enddo
      if (mfit.gt.1) then
        do j=2,mfit
          do k=1,j-1
            covar(k,j)=covar(j,k)
          enddo
        enddo
      endif
      call gaussj(covar,mfit,ncvm,beta,1,1)
      do j=1,mfit
        a(lista(j))=beta(j)
      enddo
      chisq=0.
      do i=1,ndata
        afunc (1) = 1.
        afunc (2) = x (i)
        afunc (3) = x (i) * x (i)
        afunc (4) = (x (i))**3
        sum=0.
        do j=1,ma
          sum=sum+a(j)*afunc(j)
        enddo
        chisq=chisq+(y(i)-sum)**2
      enddo
      return
      end

      subroutine gaussj(a,n,np,b,m,mp)
      implicit real*8 (a - h, o - z)
      parameter (nmax=40000)
      dimension a(np,np),b(np,mp),ipiv(nmax),indxr(nmax),indxc(nmax)
      do j=1,n
        ipiv(j)=0
      enddo
      do i=1,n
        big=0.
        do j=1,n
          if(ipiv(j).ne.1)then
            do k=1,n
              if (ipiv(k).eq.0) then
                if (abs(a(j,k)).ge.big)then
                  big=abs(a(j,k))
                  irow=j
                  icol=k
                endif
              else if (ipiv(k).gt.1) then
                pause 'singular matrix'
              endif
            enddo
          endif
        enddo
        ipiv(icol)=ipiv(icol)+1
        if (irow.ne.icol) then
          do l=1,n
            dum=a(irow,l)
            a(irow,l)=a(icol,l)
            a(icol,l)=dum
          enddo
          do l=1,m
            dum=b(irow,l)
            b(irow,l)=b(icol,l)
            b(icol,l)=dum
          enddo
        endif
        indxr(i)=irow
        indxc(i)=icol
        if (a(icol,icol).eq.0.) pause 'singular matrix.'
        pivinv=1./a(icol,icol)
        a(icol,icol)=1.
        do l=1,n
          a(icol,l)=a(icol,l)*pivinv
        enddo
        do l=1,m
          b(icol,l)=b(icol,l)*pivinv
        enddo
        do ll=1,n
          if(ll.ne.icol)then
            dum=a(ll,icol)
            a(ll,icol)=0.
            do l=1,n
              a(ll,l)=a(ll,l)-a(icol,l)*dum
            enddo
            do l=1,m
              b(ll,l)=b(ll,l)-b(icol,l)*dum
            enddo
          endif
        enddo
      enddo
      do l=n,1,-1
        if(indxr(l).ne.indxc(l))then
          do k=1,n
            dum=a(k,indxr(l))
            a(k,indxr(l))=a(k,indxc(l))
            a(k,indxc(l))=dum
          enddo
        endif
      enddo
      return
      end

      subroutine spline(x,y,n,y2)

c     Modified to always use "natural" boundary conditions

      implicit real*8 (a - h, o - z)
      parameter (maxkpno=40000)
      dimension x(n),y(n),y2(n),u(maxkpno)
      save

      y2(1)=0.
      u(1)=0.
      do i=2,n-1
        sig=(x(i)-x(i-1))/(x(i+1)-x(i-1))
        p=sig*y2(i-1)+2.
        y2(i)=(sig-1.)/p
        u(i)=(6.*((y(i+1)-y(i))/(x(i+1)-x(i))-(y(i)-y(i-1))
     *  /(x(i)-x(i-1)))/(x(i+1)-x(i-1))-sig*u(i-1))/p
      enddo
      qn=0.
      un=0.
      y2(n)=(un-qn*u(n-1))/(qn*y2(n-1)+1.)
      do k=n-1,1,-1
        y2(k)=y2(k)*y2(k+1)+u(k)
      enddo
      return
      end

      subroutine splint(xa,ya,y2a,n,np,x,y)

      implicit real*8 (a - h, o - z)
      parameter (maxpts=40000)
      dimension x (maxpts), y (maxpts)
      dimension xa(n),ya(n),y2a(n)
      save

      do i=1,np
      klo=1
      khi=n
1     if (khi-klo.gt.1) then
        k=(khi+klo)/2
        if(xa(k).gt.x(i))then
          khi=k
        else
          klo=k
        endif
      goto 1
      endif
      h=xa(khi)-xa(klo)
      if (h.eq.0.) pause 'bad xa input.'
      a=(xa(khi)-x(i))/h
      b=(x(i)-xa(klo))/h
      y(i)=a*ya(klo)+b*ya(khi)+
     * ((a**3-a)*y2a(klo)+(b**3-b)*y2a(khi))*(h**2)/6.
      end do

      return
      end
